﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Persistence;

namespace doLogic
{
    public class Users : PersistentList<User>
    {
		//Password property only contains hash of password
		public static User Login(string userName, string password)
		{
			return Users.All.Single(user => user.Login.ToLower() == userName.ToLower() && user.Password == user.PasswordHash(password));
		}

		//finds a user by username and email
		public static User RequestPasswordReset(string userName, string email)
		{
			return Users.All.Single(user => user.Login.ToLower() == userName.ToLower() && user.Email.ToLower() == email.ToLower());
		}

		//find a user by username
		public static User ByUserName(string userName)
		{
			return Users.All.Single(u => u.Login.ToLower() == userName.ToLower());
		}

		public static User AddUser(string userName, string password, string lastName, string firstName, string email)
		{
			User user = new User(userName, lastName, firstName, email);

			user._Password = "Adding new user" + email;

			//save to get the identity key
			user.Store();

			//resaving with salt on password
			user.SetPassword(password);
			user.Store();

			Users.All.Add(user);
			return user;
		}

		//loads all users into memory
        private Users()
        {
            this.Load();
        }

		//standard way to find a User or iterate through the list
        internal static Users _Users = null;
		public static Users All
		{
			get
			{
				if (_Users == null)
					_Users = new Users();
				return _Users;
			}
		}

		public static void Delete(User user)
		{
			Database.Delete(user);
			Users.All.Remove(user);
		}
    }

    [Persistable("secUsr", "ID", false)]
    public class User
    {
		public User() { }

		internal User(string userName, string lastName, string firstName, string email)
		{
			this.Login = userName;
			this.LastName = lastName;
			this.FirstName = firstName;
			this.Email = email;
		}

		private Roles _Roles = null;
		public Roles Roles
		{
			get
			{
				if (_Roles == null)
					_Roles = Roles.ForUser(this);
				return _Roles;
			}
		}

		public bool IsInRole(Role role)
		{
			return (this.Roles.Contains(role));
		}

		public bool IsInRole(int roleID)
		{
			return (this.Roles.ItemById(roleID).IsNotEmpty());
		}

		public void AddRole(Role role)
		{
			UserRole ur = new UserRole(this, role);
			Database.Store(ur);
			this.Roles.Add(role);
		}

		public void RemoveRole(Role role)
		{
			UserRole ur = new UserRole(this, role);
			Database.Delete(ur);
			this.Roles.Remove(role);
		}

		private Groups _Groups = null;
		public Groups Groups
		{
			get
			{
				if (_Groups == null)
					_Groups = Groups.ForUser(this);
				return _Groups;
			}
		}

		public bool IsInGroup(Group group)
		{
			return (this.Groups.Contains(group));
		}

		public bool IsInGroup(int groupID)
		{
			return (this.Groups.ItemById(groupID).IsNotEmpty());
		}

		public void AddGroup(Group group)
		{
			UserGroup ug = new UserGroup(this, group);
			Database.Store(ug);
			this.Groups.Add(group);
		}

		public void RemoveGroup(Group group)
		{
			UserGroup ug = new UserGroup(this, group);
			Database.Delete(ug);
			this.Groups.Remove(group);
		}
		
		public string DisplayName
        {
            get
            {
                return LastName + ", " + FirstName;
            }
        }

		internal string _Password = "";
		public string Password
		{
			get
			{
				return _Password;
			}
			set
			{
				throw new ApplicationException("To change the password, use SetPassword.");
			}
		}

		public void SetPassword(string password)
		{
			_Password = PasswordHash(password);
		}

		internal string PasswordHash(string password)
		{
			//choice of salt value is either company prefix taking responsibility for all user passwords or database administrator.
			//Byte[] bytes = Encoding.UTF8.GetBytes(password + Users.All.CreateNewItem().GetPersistenceInfo().DataSource);
			
			Byte[] bytes = Encoding.UTF8.GetBytes(password + Database.SpPrefix + this.ID.ToString());
			return BitConverter.ToString((new System.Security.Cryptography.SHA1CryptoServiceProvider()).ComputeHash(bytes)).Replace("-", "");
		}

        //persistent properties
        public int ID { get; set; }
        public string FirstName { get; set; }
        public string LastName { get; set; }
        public string Login { get; set; }
        public string Email { get; set; }
        public string Phone { get; set; }
        public bool Internal { get; set; }
        public bool Active { get; set; }
        public bool NewUser { get; set; }

		//since all users are loaded in the PersistentList<User> class (Users.All), Loads can not be done individually.  Delete is also on Users.
		public void Store()
		{
			Database.Store(this);
		}

		[Persistable("secUsrRole", "ID", false)]
		private class UserRole
		{
			public UserRole(User user, Role role)
			{
				this.usrID = user.ID;
				this.roleID = role.ID;
			}

			public int usrID { get; set; }
			public int roleID { get; set; }
		}

		[Persistable("secUsrGroup", "ID", false)]
		private class UserGroup
		{
			public UserGroup(User user, Group group)
			{
				this.usrID = user.ID;
				this.groupID = group.ID;
			}

			public int usrID { get; set; }
			public int groupID { get; set; }
		}
	}
}
